@@ -8,11 +8,10 @@
 // #include "cpl_conv.h" // for CPLFree()
 
 #include "gdallibrary/gdallibrary.h"
+#include "gdalmiscutils/gdalmiscutils.h"
 namespace gdalgeometry {
 using namespace Rcpp;
 
-
-
 // low level geometry converters from feature ----------------------------------------------
 //FIXME
 // o add cast capability?
@@ -24,21 +23,32 @@ using namespace Rcpp;
 // see here for the constants for the format variants
 // http://www.gdal.org/ogr__core_8h.html#a6716bd3399c31e7bc8b0fd94fd7d9ba6a7459e8d11fa69e89271771c8d0f265d8
 inline RawVector gdal_geometry_raw(OGRFeature *poFeature) {
-  Rcpp::RawVector raw(poFeature->GetGeometryRef()->WkbSize());
-  poFeature->GetGeometryRef()->exportToWkb(wkbNDR, &(raw[0]), wkbVariantIso);
-  return raw;
+  if (poFeature->GetGeometryRef()) {
+    Rcpp::RawVector raw(poFeature->GetGeometryRef()->WkbSize());
+    poFeature->GetGeometryRef()->exportToWkb(wkbNDR, &(raw[0]), wkbVariantIso);
+    return raw;
+    
+  } else {
+    Rcpp::RawVector raw(0); 
+    return raw; 
+  }
 }
 inline CharacterVector gdal_geometry_wkt(OGRFeature *poFeature) {
-  char *pszGEOM_WKT = NULL;
-  poFeature->GetGeometryRef()->exportToWkt(&pszGEOM_WKT, wkbVariantIso );
   CharacterVector wkt(1);
-  wkt[0] = pszGEOM_WKT;
-  CPLFree( pszGEOM_WKT );
+  if (poFeature->GetGeometryRef()) {
+    char *pszGEOM_WKT = NULL;
+    poFeature->GetGeometryRef()->exportToWkt(&pszGEOM_WKT, wkbVariantIso );
+    wkt[0] = pszGEOM_WKT;
+    CPLFree( pszGEOM_WKT );
+  } else {
+    wkt[0] = NA_STRING; 
+  }
   return wkt;
 }
 inline CharacterVector gdal_geometry_txt(OGRFeature *poFeature, CharacterVector format) {
   char *export_txt = NULL;
   CharacterVector txt(1);
+  if (poFeature->GetGeometryRef()) {
   if (format[0] == "gml") {
     export_txt = poFeature->GetGeometryRef()->exportToGML();
   }
@@ -49,30 +59,51 @@ inline CharacterVector gdal_geometry_txt(OGRFeature *poFeature, CharacterVector
     export_txt = poFeature->GetGeometryRef()->exportToKML();
   }
   txt[0] = export_txt;
+  } else {
+    txt[0] = NA_STRING; 
+  }
   CPLFree(export_txt);
   return txt;
 }
 inline IntegerVector gdal_geometry_type(OGRFeature*poFeature) {
+  OGRGeometry *poGeom = poFeature->GetGeometryRef(); 
+  
+  
+  OGRBoolean empty;
+  if (poGeom) {
+    empty = poFeature->GetGeometryRef()->IsEmpty(); 
+  } else {
+    empty = 1; 
+  }
+  
+  Rcpp::IntegerVector r_gtyp = Rcpp::IntegerVector(1);
+  if (empty) {
+    r_gtyp = NA_INTEGER; 
+  } else {
     OGRwkbGeometryType gtyp = OGR_G_GetGeometryType( poFeature->GetGeometryRef());
-    Rcpp::IntegerVector r_gtyp = Rcpp::IntegerVector(1);
-     r_gtyp[0] = (int)gtyp;
-    return r_gtyp;
+    r_gtyp[0] = (int)gtyp;
+  }
+  return r_gtyp;
 }
 inline NumericVector gdal_geometry_extent(OGRFeature *poFeature) {
-  OGREnvelope env;
-  OGR_G_GetEnvelope(poFeature->GetGeometryRef(), &env);
-  // if geometry is empty, set the envelope to undefined (otherwise all 0s)
+  
   double minx, maxx, miny, maxy;
-  if (poFeature->GetGeometryRef()->IsEmpty()) {
-    minx = NA_REAL;
-    maxx = NA_REAL;
-    miny = NA_REAL;
-    maxy = NA_REAL;
-  } else {
-    minx = env.MinX;
-    maxx = env.MaxX;
-    miny = env.MinY;
-    maxy = env.MaxY;
+  minx = NA_REAL;
+  maxx = NA_REAL;
+  miny = NA_REAL;
+  maxy = NA_REAL;
+  
+  if (poFeature->GetGeometryRef()) {
+   OGREnvelope env;
+   OGR_G_GetEnvelope(poFeature->GetGeometryRef(), &env);
+   if (poFeature->GetGeometryRef()->IsEmpty()) {
+     
+   } else {
+     minx = env.MinX;
+     maxx = env.MaxX;
+     miny = env.MinY;
+     maxy = env.MaxY;
+   }
   }
   NumericVector extent = NumericVector::create(minx, maxx, miny, maxy);
   return extent;
@@ -80,31 +111,449 @@ inline NumericVector gdal_geometry_extent(OGRFeature *poFeature) {
 
 // --------------------------------------------------------------------------------------
 
-/// READ FIDS ----------------------------------------------------------------------------
-inline NumericVector layer_read_fids_all(OGRLayer *poLayer) {
-  //double   nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  // trying to fix SQL problem 2020-10-05
-  R_xlen_t nFeature = poLayer->GetFeatureCount();
-  if (nFeature < 1) {
-    //Rprintf("force count\n");
-    nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  }
 
-  NumericVector out(nFeature);
+inline NumericVector layer_read_fids_ij(OGRLayer *poLayer, NumericVector ij) {
+  R_xlen_t st = (R_xlen_t)ij[0]; 
+  R_xlen_t en = (R_xlen_t)ij[1]; 
+  NumericVector out(en - st + 1);
+  std::fill( out.begin(), out.end(), NumericVector::get_na() );
+  OGRFeature *poFeature;
+  R_xlen_t cnt = 0;
+  R_xlen_t ii = 0;
+  
+  while(ii <= en && (poFeature = poLayer->GetNextFeature()) != NULL ) {
+    if (ii >= st) {
+      out[cnt] = (double)poFeature->GetFID();
+      cnt++;
+    }
+    ii++;
+    OGRFeature::DestroyFeature(poFeature);
+  }
+  if (cnt < out.length()) {
+    Rcpp::warning("not as many FIDs as requested");
+  }
+  return out;
+}
+
+/// layer READ FIDS ----------------------------------------------------------------------------
+inline NumericVector layer_read_fids_all(OGRLayer *poLayer) {
+  R_xlen_t nFeature = poLayer->GetFeatureCount();
+  
+  NumericVector ij(2);
+  ij[0] = 0;
+  /// if there are no features we get 0 fields allocation
+  ij[1] = (double)(nFeature - 1); 
+  
+  NumericVector out = layer_read_fids_ij(poLayer, ij);
+  return out;
+}
+
+inline NumericVector layer_read_fids_ia(OGRLayer *poLayer, NumericVector ia) {
+  
+  NumericVector out(ia.length());
   std::fill( out.begin(), out.end(), NumericVector::get_na() );
   OGRFeature *poFeature;
   R_xlen_t ii = 0;
+  R_xlen_t cnt = 0;
   while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    out[ii] = (double)poFeature->GetFID();
+    if (ii == (R_xlen_t)(ia[cnt])) {
+      out[cnt] = (double)poFeature->GetFID();
+      cnt++;
+    }
+    ii++;
+    OGRFeature::DestroyFeature(poFeature);
+  }
+  if (cnt < out.length()) {
+    Rcpp::warning("not all FIDS found");
+  }
+  return out;
+}
+
+
+inline List feature_read_geom(OGRFeature *poFeature, CharacterVector format) {
+  List out(1); 
+  // work through format
+  // FIXME: get rid of "geometry"
+  if ((format[0] == "wkb")) { // || (format[0] == "geometry")) {
+    out[0] = gdal_geometry_raw(poFeature);
+  }
+  if (format[0] == "wkt") {
+    out[0] = gdal_geometry_wkt(poFeature);
+  }
+  // FIXME: maybe call it envelope not extent
+  if (format[0] == "extent") {
+    out[0] = gdal_geometry_extent(poFeature);
+  }
+  // these are all just text variants (wkt uses a different mech)
+  if ((format[0] == "gml") || (format[0] == "json") || (format[0] == "kml")) {
+    out[0] = gdal_geometry_txt(poFeature, format);
+  }
+  if (format[0] == "type") {
+    out[0] = gdal_geometry_type(poFeature);
+  }
+  
+  return out;
+}
+/// layer READ GEOMS ----------------------------------------------------------------------------
+
+inline List layer_read_geom_ij(OGRLayer *poLayer, CharacterVector format, NumericVector ij) {
+  OGRFeature *poFeature;
+  R_xlen_t st = (R_xlen_t)ij[0]; 
+  R_xlen_t en = (R_xlen_t)ij[1]; 
+  
+  List out(en - st + 1);
+  R_xlen_t ii = 0;
+  R_xlen_t cnt = 0;
+  
+  while(ii <= en &&  (poFeature = poLayer->GetNextFeature()) != NULL ) {
+    if (ii >= st) {
+      out[cnt] = feature_read_geom(poFeature, format)[0]; 
+      cnt++;
+    }
+    OGRFeature::DestroyFeature(poFeature);
+    ii++;
+  }
+  if (cnt < out.length()) {
+    Rcpp::warning("not as many geoms as requested");
+  }
+  return out;
+}
+
+inline List layer_read_geom_all(OGRLayer *poLayer, CharacterVector format) {
+  R_xlen_t nFeature = poLayer->GetFeatureCount();
+  
+  NumericVector ij(2);
+  ij[0] = 0;
+  /// if there are no features we get 0 fields allocation
+  ij[1] = (double)(nFeature - 1); 
+  
+  List out = layer_read_geom_ij(poLayer, format, ij);
+  return out;
+}
+
+inline List layer_read_geom_ia(OGRLayer *poLayer, CharacterVector format, NumericVector ia) {
+  OGRFeature *poFeature;
+  
+  poLayer->ResetReading();
+  
+  List out(ia.length());
+  R_xlen_t ii = 0;
+  R_xlen_t cnt = 0;
+  
+  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
+    if (ii == static_cast<R_xlen_t>(ia[cnt])) {
+      out[cnt] = feature_read_geom(poFeature, format)[0]; 
+      cnt++;
+      
+    }
+    ii++;
+    OGRFeature::DestroyFeature(poFeature);
+    if (cnt == ia.length()) break; 
+  }
+  
+  return out;
+}
+
+inline List layer_read_geom_fa(OGRLayer *poLayer, CharacterVector format, NumericVector fa) {
+  OGRFeature *poFeature;
+  //poLayer->ResetReading();
+  
+  List out(fa.length());
+  
+  for (R_xlen_t ii = 0; ii < fa.length(); ii++) {
+    GIntBig feature_id = (GIntBig)fa[ii];
+    poFeature = poLayer->GetFeature(feature_id);
+    out[ii] = feature_read_geom(poFeature, format)[0]; 
+    OGRFeature::DestroyFeature(poFeature);
+  }
+  return out;
+}
+
+/// layer READ FIELDS ----------------------------------------------------------------------------
+
+// _alll uses _ij so this comes first
+inline List layer_read_fields_ij(OGRLayer *poLayer, CharacterVector fid_column_name,
+                                 NumericVector ij) {
+  
+  R_xlen_t st = (R_xlen_t)ij[0]; 
+  R_xlen_t en = (R_xlen_t)ij[1]; 
+  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
+  bool int64_as_string = false;
+  // special case if ij[1] < ij[0] 
+  R_xlen_t nfeatures = en - st + 1;
+  if (en < st) nfeatures = 0; 
+  List out = gdallibrary::allocate_fields_list(poFDefn, nfeatures, 
+                                               int64_as_string, fid_column_name);
+  OGRFeature *poFeature;
+  R_xlen_t cnt = 0;
+  R_xlen_t ii = 0;
+  int iField;
+  int bytecount; 
+  int not_NA; 
+  
+  // for OFTDateTime
+  Rcpp::Function ISOdatetime("ISOdatetime");
+  Rcpp::Function ISOdate("ISOdate");
+  while(ii <= en &&  (poFeature = poLayer->GetNextFeature()) != NULL ) {
+    if (ii >= st) {
+      for(iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
+      {
+        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
+        not_NA = poFeature->IsFieldSetAndNotNull(iField);
+        
+        /// stolen directly from sf and adapted thanks Edzer Pebesma
+        switch( poFieldDefn->GetType() ) {
+        case OFTWideString:
+        case OFTWideStringList: {
+          // don't do anything these are deprecated (and probably will cause version issues ...)
+        }
+          break;
+        case OFTString: {
+          CharacterVector cv;
+          cv = out[iField];
+          if (not_NA) {
+            cv[cnt] = poFeature->GetFieldAsString( iField );
+          } else {
+            cv[cnt] = NA_STRING;
+          }
+          
+        } break;
+        case OFTInteger: {
+          IntegerVector iv;
+          iv = out[iField];
+          if (! not_NA) {
+            iv[cnt] = NA_INTEGER;
+            break;
+          }
+          // does this need subtype handling (should allow raw instead of integer for OSTBoolean)
+          iv[cnt] = poFeature->GetFieldAsInteger( iField );
+
+        }  break;
+        case OFTTime:
+        case OFTDateTime:
+        case OFTDate: {
+          NumericVector nv;
+          nv = out[iField];
+
+          if (! not_NA) {
+            nv[cnt] = NA_REAL;
+            break;
+          }
+       
+          int Year, Month, Day, Hour, Minute, TZFlag;
+          float Second;
+          const char *tzone = "";
+         
+          poFeature->GetFieldAsDateTime(iField, &Year, &Month, &Day, &Hour, &Minute,
+                                        &Second, &TZFlag);
+          if (TZFlag == 100)
+            tzone = "UTC";
+          if (poFieldDefn->GetType() == OFTDateTime || poFieldDefn->GetType() == OFTTime) {
+            if (cnt == 0 && poFieldDefn->GetType() == OFTTime) {
+              Rcpp::warning("field of type 'OFTTime' converted to POSIXct: %s", poFDefn->GetFieldDefn(iField)->GetNameRef()); 
+            }
+            Rcpp::NumericVector ret = ISOdatetime(Year, Month, Day, Hour, Minute, Second, tzone); 
+            nv[cnt] = ret[0];
+          } else {
+            Rcpp::NumericVector ret = ISOdate(Year, Month, Day); 
+            nv[cnt] = ret[0];
+          }
+          break;
+        }
+          break;
+        case OFTInteger64: {
+          NumericVector iiv;
+          iiv = out[iField];
+          if (! not_NA) {
+            iiv[cnt] = NA_REAL;
+            break;
+          }
+          iiv[cnt] = poFeature->GetFieldAsDouble( iField );
+        }
+          break;
+        case OFTReal: {
+          NumericVector nv;
+          nv = out[iField];
+          if (! not_NA) {
+            nv[cnt] = NA_REAL;
+            break;
+          }
+          nv[cnt] = poFeature->GetFieldAsDouble( iField );
+
+        }
+          break;
+          case OFTStringList: {
+            // List slv;
+            // slv = out[iField];
+            //
+            // slv[cnt] = poFeature->GetFieldAsStringList(iField);
+            //
+          }
+            break;
+          case OFTRealList: {
+            // List rlv;
+            // rlv = out[iField];
+            // int rlcount;
+            // rlv[cnt] = poFeature->GetFieldAsDoubleList(iField, &rlcount);
+
+          }
+            break;
+          case OFTIntegerList: {
+            // List ilv;
+            // ilv = out[iField];
+            // int ilcount;
+            // ilv[cnt] = poFeature->GetFieldAsIntegerList(iField, &ilcount);
+          }
+            break;
+          case OFTInteger64List: {
+            // List iilv;
+            // iilv = out[iField];
+            // int iilcount;
+            // iilv[cnt] = poFeature->GetFieldAsInteger64List(iField, &iilcount);
+
+          }
+            break;
+
+        case OFTBinary: {
+          Rcpp::List bv;
+          bv = out[iField];
+
+          const GByte *bin = poFeature->GetFieldAsBinary(iField, &bytecount);
+          RawVector rb(bytecount);
+          for (int ib = 0; ib < bytecount; ib++) {
+            rb[ib] = bin[ib];
+          }
+          bv[cnt] = rb;
+        } break; 
+      }  // end switch
+      }
+      cnt++;
+    } // end if
+    ii++;
+    OGRFeature::DestroyFeature(poFeature);
+  } // end get next feature
+  if (cnt < nfeatures) {
+    Rcpp::warning("not as many features as requested");
+  }
+  return out;
+}
+
+
+inline List layer_read_fields_all(OGRLayer *poLayer, CharacterVector fid_column_name) {
+  R_xlen_t nFeature = poLayer->GetFeatureCount();
+
+  NumericVector ij(2);
+  ij[0] = 0;
+  /// if there are no features we get 0 fields allocation
+  ij[1] = (double)(nFeature - 1); 
+ 
+  List out = layer_read_fields_ij(poLayer, fid_column_name, ij);
+  return out;
+}
+inline List layer_read_fields_ia(OGRLayer *poLayer, CharacterVector fid_column_name,
+                                 NumericVector ia) {
+  R_xlen_t   nFeature = ia.length();
+  
+  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
+  bool int64_as_string = false;
+  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
+  
+  OGRFeature *poFeature;
+  
+  R_xlen_t ii = 0;
+  R_xlen_t cnt = 0;
+  int iField;
+  
+  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
+    
+    if (ii == static_cast<R_xlen_t>(ia[cnt])) {
+      
+      // FIXME copy the wider field support from sf (but how to make this a function)
+      for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
+      {
+        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
+        if( poFieldDefn->GetType() == OFTInteger   ) {
+          IntegerVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsInteger( iField );
+        }
+        
+        if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
+          NumericVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsDouble( iField );
+        }
+        
+        if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
+          CharacterVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsString( iField );
+          
+        }
+      }
+      
+      cnt++;
+    }
+    
     OGRFeature::DestroyFeature(poFeature);
     ii++;
   }
   return out;
 }
 
+inline List layer_read_fields_fa(OGRLayer *poLayer, CharacterVector fid_column_name,
+                                 NumericVector fa) {
+  R_xlen_t   nFeature = fa.length();
+  
+  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
+  bool int64_as_string = false;
+  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
+  
+  OGRFeature *poFeature;
+  R_xlen_t cnt = 0;
+  int iField;
+  
+  for (R_xlen_t ii = 0; ii < fa.length(); ii++) {
+    GIntBig feature_id = (GIntBig)fa[ii];
+    poFeature = poLayer->GetFeature(feature_id);
+    if (NULL == poFeature) {
+      Rcpp::warning("FID not found %i", fa[ii]);
+    } else {
+      // FIXME copy the wider field support from sf (but how to make this a function)
+      for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
+      {
+        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
+        if( poFieldDefn->GetType() == OFTInteger   ) {
+          IntegerVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsInteger( iField );
+        }
+        
+        if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
+          NumericVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsDouble( iField );
+        }
+        
+        if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
+          CharacterVector nv;
+          nv = out[iField];
+          nv[cnt] = poFeature->GetFieldAsString( iField );
+          
+        }
+      }
+      OGRFeature::DestroyFeature(poFeature);
+    }
+    cnt++;
+  }
+  return out;
+}
+
+
+// DSN read
 inline NumericVector dsn_read_fids_all(CharacterVector dsn, IntegerVector layer,
                                        CharacterVector sql, NumericVector ex) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -120,33 +569,10 @@ inline NumericVector dsn_read_fids_all(CharacterVector dsn, IntegerVector layer,
   GDALClose(poDS);
   return out;
 }
-
-inline NumericVector layer_read_fids_ij(OGRLayer *poLayer, NumericVector ij) {
-  NumericVector out((R_xlen_t)ij[1] - (R_xlen_t)ij[0] + 1);
-  std::fill( out.begin(), out.end(), NumericVector::get_na() );
-  OGRFeature *poFeature;
-  R_xlen_t cnt = 0;
-  R_xlen_t ii = 0;
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    if (ii == static_cast<R_xlen_t>(ij[0]) || (ii > static_cast<R_xlen_t>(ij[0]) && ii <= static_cast<R_xlen_t>(ij[1]))) {
-      out[cnt] = (double)poFeature->GetFID();
-      cnt++;
-    }
-    ii++;
-    OGRFeature::DestroyFeature(poFeature);
-  }
-  if (cnt < out.length()) {
-    Rcpp::warning("not as many FIDs as requested");
-  }
-
-
-  return out;
-}
-
 inline NumericVector dsn_read_fids_ij(CharacterVector dsn, IntegerVector layer,
                                       CharacterVector sql, NumericVector ex,
                                       NumericVector ij) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -163,32 +589,10 @@ inline NumericVector dsn_read_fids_ij(CharacterVector dsn, IntegerVector layer,
   return out;
 }
 
-
-inline NumericVector layer_read_fids_ia(OGRLayer *poLayer, NumericVector ia) {
-
-  NumericVector out(ia.length());
-  std::fill( out.begin(), out.end(), NumericVector::get_na() );
-  OGRFeature *poFeature;
-  R_xlen_t ii = 0;
-  R_xlen_t cnt = 0;
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    if (ii == static_cast<R_xlen_t>(ia[cnt])) {
-      out[cnt] = (double)poFeature->GetFID();
-      cnt++;
-    }
-    ii++;
-    OGRFeature::DestroyFeature(poFeature);
-  }
-  if (cnt < out.length()) {
-    Rcpp::warning("not all FIDS found");
-  }
-  return out;
-}
-
 inline NumericVector dsn_read_fids_ia(CharacterVector dsn, IntegerVector layer,
                                       CharacterVector sql, NumericVector ex,
                                       NumericVector ia) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -204,57 +608,10 @@ inline NumericVector dsn_read_fids_ia(CharacterVector dsn, IntegerVector layer,
   GDALClose(poDS);
   return out;
 }
-
-// -----------------------------------------------------------------------------
-
-
-
-
-/// READ GEOMS ----------------------------------------------------------------------------
-inline List layer_read_geom_all(OGRLayer *poLayer, CharacterVector format) {
-  OGRFeature *poFeature;
-  poLayer->ResetReading();
-  //int nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  // trying to fix SQL problem 2020-10-05
-  R_xlen_t nFeature = poLayer->GetFeatureCount();
-  if (nFeature < 1) {
-    //Rprintf("force count\n");
-    nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  }
-
-  List out(nFeature);
-  R_xlen_t ii = 0;
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    // work through format
-    // FIXME: get rid of "geometry"
-    if ((format[0] == "wkb") || (format[0] == "geometry")) {
-      out[ii] = gdal_geometry_raw(poFeature);
-    }
-    if (format[0] == "wkt") {
-      out[ii] = gdal_geometry_wkt(poFeature);
-    }
-    // FIXME: maybe call it envelope not extent
-    if (format[0] == "extent") {
-      out[ii] = gdal_geometry_extent(poFeature);
-    }
-    // these are all just text variants (wkt uses a different mech)
-    if ((format[0] == "gml") || (format[0] == "json") || (format[0] == "kml")) {
-      out[ii] = gdal_geometry_txt(poFeature, format);
-    }
-    if (format[0] == "type") {
-      out[ii] = gdal_geometry_type(poFeature);
-    }
-
-    OGRFeature::DestroyFeature(poFeature);
-    ii++;
-  }
-  return out;
-}
-
 inline List dsn_read_geom_all(CharacterVector dsn, IntegerVector layer,
                               CharacterVector sql, NumericVector ex,
                               CharacterVector format) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -270,55 +627,10 @@ inline List dsn_read_geom_all(CharacterVector dsn, IntegerVector layer,
   GDALClose(poDS);
   return out;
 }
-
-
-
-
-
-inline List layer_read_geom_ij(OGRLayer *poLayer, CharacterVector format, NumericVector ij) {
-  OGRFeature *poFeature;
-
-  poLayer->ResetReading();
-  R_xlen_t nFeature;
-  nFeature = (R_xlen_t)ij[1] - (R_xlen_t)ij[0] + 1;
-  List out(nFeature);
-  R_xlen_t ii = 0;
-  R_xlen_t cnt = 0;
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    if (ii == static_cast<R_xlen_t>(ij[0]) || (ii > static_cast<R_xlen_t>(ij[0]) && ii <= static_cast<R_xlen_t>(ij[1]))) {
-
-      // work through format
-      // FIXME: get rid of "geometry"
-      if ((format[0] == "wkb") || (format[0] == "geometry")) {
-        out[cnt] = gdal_geometry_raw(poFeature);
-      }
-      if (format[0] == "wkt") {
-        out[cnt] = gdal_geometry_wkt(poFeature);
-      }
-      // FIXME: maybe call it envelope not extent
-      if (format[0] == "extent") {
-        out[cnt] = gdal_geometry_extent(poFeature);
-      }
-      // these are all just text variants (wkt uses a different mech)
-      if ((format[0] == "gml") || (format[0] == "json") || (format[0] == "kml")) {
-        out[cnt] = gdal_geometry_txt(poFeature, format);
-      }
-      if (format[0] == "type") {
-        out[ii] = gdal_geometry_type(poFeature);
-      }
-      cnt++;
-
-    }
-    ii++;
-    OGRFeature::DestroyFeature(poFeature);
-  }
-  return out;
-}
-
 inline List dsn_read_geom_ij(CharacterVector dsn, IntegerVector layer,
                              CharacterVector sql, NumericVector ex,
                              CharacterVector format, NumericVector ij) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -336,60 +648,10 @@ inline List dsn_read_geom_ij(CharacterVector dsn, IntegerVector layer,
 }
 
 
-
-
-inline List layer_read_geom_ia(OGRLayer *poLayer, CharacterVector format, NumericVector ia) {
-  OGRFeature *poFeature;
-
-  poLayer->ResetReading();
-
-  List out(ia.length());
-  R_xlen_t ii = 0;
-  R_xlen_t cnt = 0;
-
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-    if (ii == static_cast<R_xlen_t>(ia[cnt])) {
-
-      // work through format
-      // FIXME: get rid of "geometry"
-      if ((format[0] == "wkb") || (format[0] == "geometry")) {
-        out[cnt] = gdal_geometry_raw(poFeature);
-      }
-      if (format[0] == "wkt") {
-        out[cnt] = gdal_geometry_wkt(poFeature);
-      }
-      // FIXME: maybe call it envelope not extent
-      if (format[0] == "extent") {
-        out[cnt] = gdal_geometry_extent(poFeature);
-      }
-      // these are all just text variants (wkt uses a different mech)
-      if ((format[0] == "gml") || (format[0] == "json") || (format[0] == "kml")) {
-        out[cnt] = gdal_geometry_txt(poFeature, format);
-      }
-      if (format[0] == "type") {
-        out[ii] = gdal_geometry_type(poFeature);
-      }
-      cnt++;
-
-    }
-    ii++;
-    OGRFeature::DestroyFeature(poFeature);
-  }
-  // should we do this? normalize output to avoid idx greater than available?
-  // if (cnt < ia.length()) {
-  //   List out2(cnt);
-  //   for (int ii = 0; ii < cnt; ii++) {
-  //     out2[ii] = out[ii];
-  //   }
-  //   return out2;
-  // }
-  return out;
-}
-
 inline List dsn_read_geom_ia(CharacterVector dsn, IntegerVector layer,
                              CharacterVector sql, NumericVector ex,
                              CharacterVector format, NumericVector ia) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -406,45 +668,10 @@ inline List dsn_read_geom_ia(CharacterVector dsn, IntegerVector layer,
   return out;
 }
 
-
-
-inline List layer_read_geom_fa(OGRLayer *poLayer, CharacterVector format, NumericVector fa) {
-  OGRFeature *poFeature;
-  //poLayer->ResetReading();
-
-  List out(fa.length());
-
-  for (R_xlen_t ii = 0; ii < fa.length(); ii++) {
-    GIntBig feature_id = (GIntBig)fa[ii];
-    poFeature = poLayer->GetFeature(feature_id);
-    // work through format
-    // FIXME: get rid of "geometry"
-    if ((format[0] == "wkb") || (format[0] == "geometry")) {
-      out[ii] = gdal_geometry_raw(poFeature);
-    }
-    if (format[0] == "wkt") {
-      out[ii] = gdal_geometry_wkt(poFeature);
-    }
-    // FIXME: maybe call it envelope not extent
-    if (format[0] == "extent") {
-      out[ii] = gdal_geometry_extent(poFeature);
-    }
-    // these are all just text variants (wkt uses a different mech)
-    if ((format[0] == "gml") || (format[0] == "json") || (format[0] == "kml")) {
-      out[ii] = gdal_geometry_txt(poFeature, format);
-    }
-    if (format[0] == "type") {
-      out[ii] = gdal_geometry_type(poFeature);
-    }
-    OGRFeature::DestroyFeature(poFeature);
-  }
-  return out;
-}
-
 inline List dsn_read_geom_fa(CharacterVector dsn, IntegerVector layer,
                              CharacterVector sql, NumericVector ex,
                              CharacterVector format, NumericVector fa) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -461,71 +688,22 @@ inline List dsn_read_geom_fa(CharacterVector dsn, IntegerVector layer,
   return out;
 }
 
-/// --------------------------------------------------------------------------------------
-
-
-/// READ FIELDS ----------------------------------------------------------------------------
-inline List layer_read_fields_all(OGRLayer *poLayer, CharacterVector fid_column_name) {
-  //double   nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  // trying to fix SQL problem 2020-10-05
-  R_xlen_t nFeature = poLayer->GetFeatureCount();
-  if (nFeature < 0) {
-    //Rprintf("force count\n");
-    nFeature = gdallibrary::force_layer_feature_count(poLayer);
-  }
-
-  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
-  bool int64_as_string = false;
-  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
-
-  OGRFeature *poFeature;
-  R_xlen_t ii = 0;
-  int iField;
-
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-
-    // FIXME copy the wider field support from sf (but how to make this a function)
-    for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
-    {
-      OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
-      if( poFieldDefn->GetType() == OFTInteger   ) {
-        IntegerVector nv;
-        nv = out[iField];
-        nv[ii] = poFeature->GetFieldAsInteger( iField );
-      }
-
-      if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
-        NumericVector nv;
-        nv = out[iField];
-        nv[ii] = poFeature->GetFieldAsDouble( iField );
-      }
-
-      if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
-        CharacterVector nv;
-        nv = out[iField];
-        nv[ii] = poFeature->GetFieldAsString( iField );
-
-      }
-    }
-
-    OGRFeature::DestroyFeature(poFeature);
-    ii++;
-  }
-  return out;
-}
-
+/// dsn READ FIELDS ----------------------------------------------------------------------------
 inline List dsn_read_fields_all(CharacterVector dsn, IntegerVector layer,
                                 CharacterVector sql, NumericVector ex,
                                 CharacterVector fid_column_name) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
+  
   if( poDS == NULL )
   {
     Rcpp::stop("Open failed.\n");
   }
   OGRLayer *poLayer = gdallibrary::gdal_layer(poDS, layer, sql, ex);
+  
   List out = layer_read_fields_all(poLayer, fid_column_name);
+  
   // clean up if SQL was used https://www.gdal.org/classGDALDataset.html#ab2c2b105b8f76a279e6a53b9b4a182e0
   if (sql[0] != "") {
     poDS->ReleaseResultSet(poLayer);
@@ -534,63 +712,10 @@ inline List dsn_read_fields_all(CharacterVector dsn, IntegerVector layer,
   return out;
 }
 
-
-inline List layer_read_fields_ij(OGRLayer *poLayer, CharacterVector fid_column_name,
-                                 NumericVector ij) {
-  R_xlen_t   nFeature;
-  nFeature = (R_xlen_t)ij[1] - (R_xlen_t)ij[0] + 1;
-
-  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
-  bool int64_as_string = false;
-  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
-
-  OGRFeature *poFeature;
-
-  R_xlen_t ii = 0;
-  R_xlen_t cnt = 0;
-  int iField;
-
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-
-    if (ii == static_cast<R_xlen_t>(ij[0]) || (ii > static_cast<R_xlen_t>(ij[0]) && ii <= static_cast<R_xlen_t>(ij[1]))) {
-
-      // FIXME copy the wider field support from sf (but how to make this a function)
-      for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
-      {
-        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
-        if( poFieldDefn->GetType() == OFTInteger   ) {
-          IntegerVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsInteger( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
-          NumericVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsDouble( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
-          CharacterVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsString( iField );
-
-        }
-      }
-
-      cnt++;
-    }
-
-    OGRFeature::DestroyFeature(poFeature);
-    ii++;
-  }
-  return out;
-}
-
 inline List dsn_read_fields_ij(CharacterVector dsn, IntegerVector layer,
                                CharacterVector sql, NumericVector ex,
                                CharacterVector fid_column_name, NumericVector ij) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -606,64 +731,10 @@ inline List dsn_read_fields_ij(CharacterVector dsn, IntegerVector layer,
   GDALClose(poDS);
   return out;
 }
-// -----------------------------------------------------------------------------------------
-
-
-inline List layer_read_fields_ia(OGRLayer *poLayer, CharacterVector fid_column_name,
-                                 NumericVector ia) {
-  R_xlen_t   nFeature = ia.length();
-
-  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
-  bool int64_as_string = false;
-  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
-
-  OGRFeature *poFeature;
-
-  R_xlen_t ii = 0;
-  R_xlen_t cnt = 0;
-  int iField;
-
-  while( (poFeature = poLayer->GetNextFeature()) != NULL ) {
-
-    if (ii == static_cast<R_xlen_t>(ia[cnt])) {
-
-      // FIXME copy the wider field support from sf (but how to make this a function)
-      for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
-      {
-        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
-        if( poFieldDefn->GetType() == OFTInteger   ) {
-          IntegerVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsInteger( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
-          NumericVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsDouble( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
-          CharacterVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsString( iField );
-
-        }
-      }
-
-      cnt++;
-    }
-
-    OGRFeature::DestroyFeature(poFeature);
-    ii++;
-  }
-  return out;
-}
-
 inline List dsn_read_fields_ia(CharacterVector dsn, IntegerVector layer,
                                CharacterVector sql, NumericVector ex,
                                CharacterVector fid_column_name, NumericVector ia) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
@@ -680,61 +751,10 @@ inline List dsn_read_fields_ia(CharacterVector dsn, IntegerVector layer,
   return out;
 }
 
-
-
-
-inline List layer_read_fields_fa(OGRLayer *poLayer, CharacterVector fid_column_name,
-                                 NumericVector fa) {
-  R_xlen_t   nFeature = fa.length();
-
-  OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
-  bool int64_as_string = false;
-  List out = gdallibrary::allocate_fields_list(poFDefn, nFeature, int64_as_string, fid_column_name);
-
-  OGRFeature *poFeature;
-  R_xlen_t cnt = 0;
-  int iField;
-
-  for (R_xlen_t ii = 0; ii < fa.length(); ii++) {
-    GIntBig feature_id = (GIntBig)fa[ii];
-    poFeature = poLayer->GetFeature(feature_id);
-    if (NULL == poFeature) {
-      Rcpp::warning("FID not found %i", fa[ii]);
-    } else {
-      // FIXME copy the wider field support from sf (but how to make this a function)
-      for( iField = 0; iField < poFDefn->GetFieldCount(); iField++ )
-      {
-        OGRFieldDefn *poFieldDefn = poFDefn->GetFieldDefn( iField );
-        if( poFieldDefn->GetType() == OFTInteger   ) {
-          IntegerVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsInteger( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTReal || poFieldDefn->GetType() == OFTInteger64) {
-          NumericVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsDouble( iField );
-        }
-
-        if( poFieldDefn->GetType() == OFTString || poFieldDefn->GetType() == OFTDate || poFieldDefn->GetType() == OFTTime || poFieldDefn->GetType() == OFTDateTime) {
-          CharacterVector nv;
-          nv = out[iField];
-          nv[cnt] = poFeature->GetFieldAsString( iField );
-
-        }
-      }
-      OGRFeature::DestroyFeature(poFeature);
-    }
-    cnt++;
-  }
-  return out;
-}
-
 inline List dsn_read_fields_fa(CharacterVector dsn, IntegerVector layer,
                                CharacterVector sql, NumericVector ex,
                                CharacterVector fid_column_name, NumericVector fa) {
-
+  
   GDALDataset       *poDS;
   poDS = (GDALDataset*) GDALOpenEx(dsn[0], GDAL_OF_VECTOR, NULL, NULL, NULL );
   if( poDS == NULL )
