@@ -1,4 +1,4 @@
 library(testthat)
 library(vapour)
 
-test_check("vapour")
+#test_check("vapour")
